const pokidexVersion = "0.0.4";

/*
 * (setq js2-additional-externs (append '("roma" "PokeData" "MoveData" "CompletionHandler""CompletionSourceKeys") js2-additional-externs))
 *
 */

function message(str) {
  let echoArea = document.getElementById("echo-area");
  if (typeof str === "string") {
    echoArea.textContent = str;
  }
  else {
    echoArea.innerHTML = "";
    echoArea.appendChild(str);
  }
}

function messageHTML(str) {
  document.getElementById("echo-area").innerHTML = str;
}

function makeInput(type, name, opts) {
  let input = document.createElement("input");
  input.type = type;
  input.name = name;
  if (opts) {
    for (let i in opts) {
      if (opts[i] != undefined) {
        input[i] = opts[i];
      }
    }
  }
  return input;
}


function makeInputText(name, size, placeholder, completionSource) {
  let input = makeInput("text", name, {size: size, placeholder: placeholder});
  if (completionSource) {
    CompletionHandler.on(input, completionSource);
  }
  return input;
}

function makeInputNumber(name, size, min, max, init) {
  let input = makeInput("number", name, {min: min, max: max});
  input.style.width = size + 3 + "ex";
  input.value = init;
  return input;
}

function makeInputButton(name, value, action) {
  return makeInput("button", name, {value: value, onclick: action});
}

function romaWithSuffix(str) {
  return new RegExp("^" + roma(str, true));
}

function createElementWithText(tag, text) {
  let node = document.createElement(tag);
  node.textContent = text;
  return node;
}

function makeDiv(klass, ...args) {
  let div = document.createElement("div");
  if (klass) {
    div.className = klass;
  }
  for (let e of args) {
    div.appendChild(e);
  }
  return div;
}

function makeBlock(name, htext, text) {
  let div = document.createElement("div");
  let hline = createElementWithText("h3", htext);
  div.className = name;
  div.appendChild(hline);
  div.appendChild(createElementWithText("pre", text));
  return div;
}

function makeImage(name, path, title) {
  let img = document.createElement("img");
  img.src = path;
  img.className = name;
  img.title = title;
  img.addEventListener("load", function () {
    img.height = img.naturalHeight * 2;
    img.width = img.naturalWidth * 2;
  }, false);
  return img;
}

function resultCloser(node) {
  let div = document.createElement("div");
  div.appendChild(makeInputButton(null, "閉じる", ev => node.parentNode.removeChild(node)));
  div.className = "close";
  return div;
}

function insertResult(node, after) {
  let result = document.getElementById("result");
  result.insertBefore(node, after ? after.nextSibling : result.firstChild);
}

function MessageException(mes) {
  this.message = mes;
};

function getPokemonFromNameOrID(key) {
  try {
    return PokeData.index[key] || PokeData.fromID(parseInt(key));
  }
  catch (e) {
    if (e instanceof RangeError) {
      throw new MessageException("エラー: 存在しない種族名またはID");
    }
    else {
      throw e;
    }
  }
}

function showPokidex(name, after) {
  let p = getPokemonFromNameOrID(name);
  let page = document.createElement("article");
  page.appendChild(resultCloser(page));

  function formatLV(n) {
    return n < 10 ? " " + n : "" + n;
  }
  function formatTMNumber(n) {
    return n < 10 ? " " + n :
      n <= 50 ? "" + n :
      n <= 57 ? "H" + (n - 50) : "--";
  }
  function formatEV1(triple) {
    let [id, type, value] = triple;
    let str = "→ No. " + id + " " + PokeData.fromID(id).name + "(";
    if (type === 1) {
      str += "Lv " + value;
    }
    else if (type === 2) {
      str += "石";
    }
    else if (type === 3) {
      str += "通信";
    }
    else if (type === 4) {
      str += "なつき";
    }
    else if (type === 5) {
      str += "その他";
    }
    return str + ")";
  }
  function formatEV(evlist) {
    return evlist.map(formatEV1).join("\n");
  }
  function formatLearnings(alist) {
     return alist.map(e => "Lv " + formatLV(e[0]) + " : " + MoveData.fromID(e[1]).name).join("\n") + "\n";
  }
  function formatTM(ary, tmlist) {
     return ary.map(i => formatTMNumber(i) + " : " + tmlist[i].name).join("\n") + "\n";
  }
  let sp = [p.h, p.a, p.b, p.c, p.d, p.s];
  let outline =
        makeBlock(
          "pokidex-block",
          "No. " + p.no + " " + p.name,
          [sp.join("-") + " : " + sp.reduce((r, i) => r + i),
           PokeData.type[p.type1] + (p.type2 ? "/" + PokeData.type[p.type2] : ""),
           p.female ? "♀率: " + p.female + "/16" : "性別不明",
           "タマゴグループ: " + p.eggGroup.map(i => PokeData.eggGroup[i]).join("/"),
          ].join("\n"));
  if (p.evList) {
    for (let triple of p.evList) {
      outline.appendChild(makeTextLinkToPokidex(PokeData.fromID(triple[0]).name, page, formatEV1(triple)));
      outline.appendChild(document.createElement("br"));
    }
  }
  if (p.evFrom) {
    outline.appendChild(makeTextLinkToPokidex(p.evFrom, page, "← No. " + p.evFrom + " " + PokeData.fromID(p.evFrom).name));

  }

  page.appendChild(
    makeDiv(
      "pokidex-box",
      outline,
      makeDiv(
        "pokidex-block pokidex-icon-block",
        makeImage("pokidex-icon", "../img/crystal/" + p.no + ".png", "通常"),
        makeImage("pokidex-icon", "../img/crystal-shiny/" + p.no + ".png", "色違い"))));
  page.appendChild(
    makeDiv(
      "pokidex-box",
      makeBlock(
        "pokidex-block",
        "クリスタル版で覚える技",
        formatLearnings(p.crystalLearnings)),
      makeBlock(
        "pokidex-block",
        "ピカチュウ版で覚える技",
        formatLearnings(p.pikaLearnings))));
  page.appendChild(
    makeDiv(
      "pokidex-box",
      makeBlock(
        "pokidex-block",
        "わざマシン",
        formatTM(p.TMMoves, MoveData.TM)),
      makeBlock(
        "pokidex-block",
        "旧作わざマシン",
        formatTM(p.oldTMMoves, MoveData.oldTM))));
  page.appendChild(
    makeDiv(
      "pokidex-box",
      makeBlock(
        "pokidex-block",
        "タマゴわざ",
        p.origin().eggMoves.map(id => MoveData.fromID(id).name).join("\n"))));

  insertResult(page, after);
}


/*
 * 技検索
 */

function takeWhile(ary, f) {
  let r = [];
  for (let e of ary) {
    if (f(e)) {
      r.push(e);
    }
    else {
      break;
    }
  }
  return r;
}

function dropWhile(ary, f) {
  for (let i = 0; i < ary.length; i++) {
    if (!f(i)) {
      return ary.slice(i);
    }
  }
  return [];
}

function arrayIntersection(nd, er) {
  return nd.filter(i => er.indexOf(i) >= 0);
}
function arrayDifference(nd, er) {
  return nd.filter(i => er.indexOf(i) < 0);
}

function filterLearningsBySelf(requirements, alist, level = 100) {
  return arrayDifference(requirements,
                         takeWhile(alist, t => t[0] <= level).
                         map(t => MoveData.fromID(t[1])));
}

function filterLearningsInCrystal(requirements, poke, level = 100) {
  return filterLearningsBySelf(requirements, poke.crystalLearnings, level);
}

function filterLearningsInPika(requirements, poke, level = 100) {
  return filterLearningsBySelf(requirements, poke.pikaLearnings, level);
}

function filterLearningsByTM(requirements, poke) {
  return arrayDifference(requirements,
                         poke.TMMoves.map(i => MoveData.TM[i]));
}

function filterLearningsByOldTM(requirements, poke) {
  return arrayDifference(requirements,
                         poke.oldTMMoves.map(i => MoveData.oldTM[i]));
}

function filterLeaningsFinalForm(requirements, poke, level = 100) {
  return filterLearningsInCrystal(filterLearningsByTM(requirements, poke, level),
                                  poke,
                                  level);
}

function isOldPoke(poke) {
  return poke.no < 152;
}

function isOldMove(move) {
  return move.no < 167;
}

// まだ使ってない関数
// 進化後のポケモンから進化前のポケモンの進化レベルを返す
// 進化しないポケモンとその他の進化の区別は？
function evolvedLevel(poke) {
  let triple;
  // poke.evFrom.evList この辺おかしくない？
  if (poke.evFrom && (triple = poke.evFrom.evList.find(triple => triple[0] === poke.no))) {
    let evType = triple[1];
    return (evType === 1) ? triple[2] :
      (evType === 4 || evType === 5) ? 2 :
      null;
  }
  else {
    return null;
  }
}

// 覚える方法を記録しながら探索したいので設計しなおしたい
// 同時に覚えられるかどうか
function pokeLearningsMovesCompatibly(poke, moves, level = 100) {
  // 最終形態で覚えられる技を除外
  let rest = filterLeaningsFinalForm(moves, poke, level);

  // *** ここから適当実装 ***
  let origin = poke.origin();
  // タマゴ技を両立できるかどうかは一旦無視する
  rest = arrayDifference(rest, origin.eggMoves.map(id => MoveData.fromID(id)));

  let p;
  for (let id = poke.no; id; id = p.evFrom) {
    p = PokeData.fromID(id);
    // 本当はレベル考えないといけない
    rest = filterLeaningsFinalForm(rest, p, level);
    // 前作も考えないといけない
    rest = filterLearningsInPika(rest, p, level);
    rest = filterLearningsByOldTM(rest, p);
  }
  // *** ここまで適当実装 ***

  return rest.length <= 0;
  // 取れる行動は 進化前 通信 タマゴ
  // タマゴわざの両立とレベル技遺伝 スケッチ
}

function iota(len, init = 0, step = 1) {
  let r = new Array(len);
  for (let i = 0; i < len; i++) {
    r[i] = init;
    init += step;
  }
  return r;
}

function makeTextLinkToPokidex(name, after, text) {
  let e = document.createElement("a");
  e.href = location.href;
  e.addEventListener("click", ev => (ev.preventDefault(), showPokidex(name, after)), false);
  e.textContent = text != null ? text : name;
  return e;
}

function search() {
  let pokidex = document.pokidex;
  let moves = iota(4, 1).map(i => pokidex["move" + i].value).map(name => MoveData.index[name]).filter(x => x);
  let lv = pokidex.lv.value || 100;
  let typeNames = [pokidex.type1.value, pokidex.type2.value].filter(
    x => x
  );
  let types = typeNames.map(
    typeName => PokeData.type.indexOf(typeName)
  );
  let pokes = PokeData.raw.filter(
    poke => pokeLearningsMovesCompatibly(poke, moves, lv)
  ).filter(
    poke => types.every(
      type => type === poke.type1 || type === poke.type2
    )
  );

  let page = document.createElement("article");
  page.appendChild(resultCloser(page));
  page.appendChild(createElementWithText("p", "レベル " + lv + " までに わざ " + moves.map(move => move.name).join(" と ") + " を覚える" + (types.length ? " " + typeNames.join("/") + " タイプ " : "") + "ポケモンの検索結果(仮)"));

  let ul = document.createElement("ul");
  for (let link of pokes.map(poke => makeTextLinkToPokidex(poke.name, page))) {
    let li = document.createElement("li");
    li.appendChild(link);
    ul.appendChild(li);
  }
  page.appendChild(ul);

  insertResult(page);
}

function scrollDown() {
  let nodes = document.getElementById("result").childNodes;
  window.scrollTo(0, Array.prototype.map.call(nodes, node => node.offsetTop).find(top => top > window.scrollY) || window.scrollMaxY);
}

function scrollUp() {
  let nodes = document.getElementById("result").childNodes;
  window.scrollTo(0, Array.prototype.map.call(nodes, node => node.offsetTop).reverse().find(top => top < window.scrollY) || window.scrollMinY);
}

function makeElement(name, attr, ...args) {
  let node = document.createElement(name);
  if (attr) {
    for (let i in attr) {
      node[i] = attr[i];
    }
  }
  for (let e of args) {
    node.appendChild(e);
  }
  return node;
}


function makeInputBox(id, arg, ...args) {
  let node = document.createElement("div");
  node.id = id;
  node.className = "input-box";
  if (arg) {
    node.appendChild(arg);
  }
  for (let e of args) {
    node.appendChild(document.createElement("br"));
    node.appendChild(e);
  }
  return node;
}

function withErrorHandler(f) {
  return function (...args) {
    try {
      return f.apply(null, args);
    } catch (er) {
      if (er instanceof MessageException) {
        message(er.message);
      }
      else {
        messageHTML("<pre>" + er.stack + "</pre>");
        throw er;
      }
    }
  };
}

addEventListener("DOMContentLoaded", function (ev) {
  try {
    message("initializing...");
    document.title += " " + pokidexVersion;
    CompletionHandler.prototype.isCompletionEvent = function (ev) {
      return ev.key === " " || ev.key === "Enter";
    };
    CompletionHandler.prototype.show = function (result) {
      let ary = result.candidates;
      if (ary.length == 0) {
        message("[No match]");
      }
      else if (ary.length <= 99) {
        let p = document.createElement("p");
        p.innerHTML = ary.sort().join("<br />");
        let columnCount = ary.length < 30 ? Math.ceil(ary.length / 6) : "auto";
        p.style.MozColumns = columnCount + " 10em";
        message(p);
      }
      else {
        message("[" + ary.length + " possibilities]");
      }
    };
    CompletionHandler.prototype.clear = _ => message("");

    {
      let dform = document.pokidex.direct,
          CSPokeData = new CompletionSourceKeys(PokeData.index, romaWithSuffix),
          show = withErrorHandler(_ => showPokidex(document.pokidex.poke.value));
      dform.appendChild(
        makeInputBox("input-direct",
                     makeInputText("poke", 7, "ポケモン", CSPokeData)));
      dform.appendChild(document.createElement("br"));
      dform.appendChild(makeInputButton("show", "表示", show));
      dform.addEventListener("keypress", function (ev) {
        if (ev.key === "Enter") {
          // うーん
          if (!CSPokeData.exists(ev.target.value)) {
            let result = CSPokeData.complete(ev.target.value);
            if (result.value) {
              ev.target.value = result.value;
            }
          }
          show();
          message("");
          ev.target.blur();
          ev.preventDefault();
        }
      }, false);
    }
    {
      let sform = document.pokidex.search,
          searchWithHandler = withErrorHandler(search),
          CSType = new CompletionSourceArray(PokeData.type, romaWithSuffix),
          CSMoveData = new CompletionSourceKeys(MoveData.index, romaWithSuffix),
          inputBoxes = makeDiv("input-boxes");
      sform.appendChild(inputBoxes);
      inputBoxes.appendChild(
        makeInputBox("input-types",
                     makeInputNumber("lv", 3, 2, 100, 55),
                     makeInputText("type1", 4, "タイプ1", CSType),
                     makeInputText("type2", 4, "タイプ2", CSType)));
      inputBoxes.appendChild(
        makeInputBox("input-moves",
                     makeInputText("move1", 10, "わざ", CSMoveData),
                     makeInputText("move2", 10, "わざ", CSMoveData),
                     makeInputText("move3", 10, "わざ", CSMoveData),
                     makeInputText("move4", 10, "わざ", CSMoveData)));
      sform.appendChild(document.createElement("br"));
      sform.appendChild(makeInputButton("search", "検索", searchWithHandler));
      sform.addEventListener("keypress", function (ev) {
        // この辺も CompletionHandler でまとまらないかな
        if (ev.key === "Enter") {
          searchWithHandler();
          message("");
          ev.target.blur();
          ev.preventDefault();
        }
      }, false);
    }

    messageHTML("POKiDEX " + pokidexVersion + "<div>クリスタル版ポケモン図鑑、タイプ・技検索付き(AND検索のみ、両立可否判断無し)。xpd風入力補完あり。検索結果や図鑑を並べて表示できる(明示的に閉じなければ閉じない)UIが特徴。Firefox推奨、Chromeでも一応動くはず。</div>" + "<pre>0.0.4\n  Chrome対応\n0.0.3\n  進化後にしか使えない技マシンを検索できなかったので修正\n0.0.2\n  Firefox44に対応した\n  とりあえずタマゴ技、旧作技を検索に全部出るようにした。両立は考えていない</pre>");
  }
  catch (e) {
    console.log(e);
  }
}, false);
